#include <sstream>
#include <iostream>
#include <string>
#include "rtree.h"
#include "conjclasses.h"



// hack: string(n, 'A' + i) doesn't work
static inline string dup_char(int n, char c)
{
  return string(n,c);
}


// Compute a representative of the conjugacy class corresponding
// to the given rtree.  Returns a string giving a product of
// generators A..Z and their inverses a..z.

string class_rep(const rtree& rt) {
  assert(isprime(rt.r)); // for now
  int n = rt.get_height();
  int r = rt.r;
  int d = rt.get_d();
  if (n == 0)
    return dup_char(rt.root, 'A' + 0);
  else
    {
      string s;

      if (rt.root == 0) {
	for (int i = 0; i < r; i++) {
	  // look at kids.  Conjugate each to shift it into position.
	  s += dup_char(i, 'A' + n) 
	    + class_rep(*rt.kids[i]) 
	    + dup_char(i * (r-1), 'A' + n); // inverse
	  // really dup_char(i, 'a' + n);
	  // for non-prime case we'd have something
	  // involving dup_char((r/d)*root % r, 'A' + n)
	}
      } else {
	assert(d == r);
	// all kids are the same.  Take one copy, in first position,
	// and z^k is tacked on the back
	s = class_rep(*rt.kids[0]) + dup_char(rt.root, 'A' + n);
	// non-prime case again will have r/d
      }
      return s;
    }
}

bint class_size(const rtree& rt) {
  assert(rt.r == 2);
  if (rt.get_height() == 0) // i.e. an element of W_1,r
    return 1;
  if (rt.root == 0) {
    const rtree kid0 = *(rt.kids[0]), kid1 = *(rt.kids[1]);
    if (kid0 == kid1) {
      bint k0size = class_size(kid0);
      return k0size * k0size;
    } else {
      return 2 * class_size(kid0) * class_size(kid1);
    }
  } else {
    assert(rt.root == 1);
    // all kids are same
    return class_size(*rt.kids[0]) * wr_order(rt.get_height(), rt.r);
  }
}

static string gen_to_cycle(char c) {
  ostringstream ost;
  int n = c - 'A';
  int ts = (1 << n); // leaves on this subtree
  for (int i = 1; i <= ts; i++) {
    ost << "(" << i << "," << (i + ts) << ")";
  }
  return ost.str();
}

string gens_to_cycle(const string& gens) {
  ostringstream ost;
  for (string::const_iterator i = gens.begin(); i != gens.end(); ++i) {
    if (i != gens.begin())
      ost << "*";
    ost << gen_to_cycle(*i);
  }
  return ost.str();
}
