#ifndef RTREE_H
#define RTREE_H

#include <assert.h>

#include <string>
#include <iostream>
#include <vector>
#include <deque>
#include <set>

#include "wreath.h"

using namespace std;

#include <gmpxx.h>

inline ostream & operator<<(ostream &o, mpz_srcptr z)
{
  char *s = mpz_get_str(NULL, 10, z);
  ostream& r = (o << s);
  free(s);
  return r;
}

template<class It> ostream& dump_range(It i, It end, ostream& os)
{
  for (; i != end; ++i)
    os << *i << ' ';
  os << endl;
  return os;
}  

template<class Cont> ostream& dump(const Cont& c, ostream& os)
{
  return dump_range(c.begin(), c.end(), os);
}
      


// number of r-trees

typedef mpz_class bint;

inline bint bpow(const bint& b, unsigned int e)
{
  bint r;
  mpz_pow_ui(r.get_mpz_t(), b.get_mpz_t(), e);
  return r;
}

bint count_rtrees(int n, int r);

// rtree stuff

struct rtree {
  int r;
  vector<rtree *> kids;
  mutable int d;
  int root;
  mutable bint dim;

  rtree (int r_) : r(r_), kids(r_, (rtree *)0), d(0), root(0), dim(0) { };
  
  bool isleaf () const { return !kids[0]; }

  void makeleaf (int root_) { assert(isleaf()); root = root_; dim = 1; }

  friend bool rpequal(const rtree *a, const rtree *b);
  int get_d () const;
  const bint& get_dim () const;
  ostream& print(ostream&) const;
  ostream& print_parsetree (ostream&) const;
  int get_height() const {
    if (isleaf())
      return 0;
    else
      return kids[0]->get_height() + 1;
  }
};

bool rpequal (const rtree* a, const rtree* b);

inline bool operator== (const rtree& a, const rtree& b)
{
  return rpequal(&a, &b);
}

inline bool operator!= (const rtree& a, const rtree& b)
{
  return !(a == b);
}

inline ostream& operator<< (ostream& os, const rtree& rt) {
  return rt.print(os);
}

extern const int max_depth;

extern vector<rtree> all_rtrees[];

extern int all_r; // all_rtrees depends on r, so it only makes sense for r to be fixed

void clear_all_rtrees(void);

void generate_all_rtrees(int depth, int r);

bint wr_order(int n, int r);

#endif
